#include <config.h>
#include <iostream>
#include <sstream>
#include <sstream>
#include <fstream>
#include <cstdlib>
#include <unistd.h>

#include "util.h"
#include "yacts.h"

// this should be ported
#include <unistd.h>
#define WAIT usleep(5000)

using namespace std;

void loadScript(istream& script) {
  yacts ex;
  ex.init(script);
};

extern "C" {
  void _stdcall print (int n, int* a) {
    cerr << "printing: ";
    for (int i=0;i<n;i++) cerr << a[i] << " ";
    cerr << endl;
  }
  
  void printarr(int r, int* s, double* y) {
    cerr << r << ", {";
    int n=1;
    for (int i=0;i<r;i++) {
      n*=s[i];
      cerr << s[i];
      if (i<r-1) cerr << ", ";
    };
    cerr << "}, {";
    for (int i=0;i<n;i++) {
      n*=y[i];
      cerr << y[i];
      if (i<n-1) cerr << ", ";
    };
    cerr << "}";
  };

  int _stdcall mf (int ry, int* sy, double* y, 
		   int rr, int* sr, double* r) {
    cerr << "mf (";
    printarr(ry, sy, y);
    cerr << ")" << endl;
    int n=1;
    for (int i=0;i<rr;i++) n*=sr[i];
    for (int i=0;i<n;i++) r[i]=i;
    return 0;
  };

  int _stdcall df (int rx, int* sx, double* x,
		   int ry, int* sy, double* y, 
		   int rr, int* sr, double* r) {
    cerr << "df (";
    printarr(rx, sx, x);
    cerr << ", ";
    printarr(ry, sy, y);
    cerr << ")" << endl;
    int n=1;
    for (int i=0;i<rr;i++) n*=sr[i];
    for (int i=0;i<n;i++) r[i]=i;
    return 0;
  };
}

extern bool PROGRESS;

int main (int argc, char **argv) {
  bool error=false;
  bool opt_h=false;
  bool opt_i=false;
  bool opt_p=false;
  int  nc=0,nj=1,fn=-1;
  bool opt_f=false;
  bool opt_s=false;
  double endTime=-1;
  string out="OUT";

  int c;
  while (((c = getopt (argc, argv, "hispfvc:j:n:o:")) != -1)&&!error)
    switch (c) {
    case 'h':
      opt_h = true;
      break;
    case 'i':
      opt_i = true;
      break;
    case 's':
      opt_s = true;
      break;
    case 'p':
      opt_p = true;
      break;
    case 'f':
      opt_f = true;
      break;
    case 'v':
      PROGRESS=true;
      break;
    case 'c':
      error=parsepositiveint(optarg,nc);
      break;
    case 'j':
      error=parsepositiveint(optarg,nj);
      break;
    case 'n':
      error=parsepositiveint(optarg,fn);
      break;
    case 'o':
      out=optarg;
      error=false;
      break;
    case 'T':
      error=parsepositivedouble(optarg,endTime);
    case '?':
      error=true;
    default:
      abort ();
    }
  if ((!error)&&(error=((opt_h+opt_i+opt_s+opt_p)>1))) {
    cerr << "Only one of -h, -i, -p, -s may be specified at the same time." << endl;
  }
  if ((!error)&&(error=(nc>=nj))) {
    cerr << "Current job number must be less than total amount of jobs," 
	 <<" but "<< nc << ">=" << nj << " ." << endl;
  }
  if ((!error)&&(error=((!opt_p)&&(nj>1)))) {
    cerr << "Parallel processing can be used only when simulation is" 
	 <<" inactive, no "<< nj <<" jobs." << endl;
  }
  if ((!error)&&(error=((!opt_p)&&(fn>0)))) {
    cerr << "Number of frames can only be set when simulation is" 
	 <<" inactive, no "<< fn <<" frames." << endl;
  }
  if ((!error)&&(error=((opt_p)&&(endTime>0)))) {
    cerr << "Can't limit simulation time, when simulation is not running." 
	 << endl;
  }
  if ((!error)&&(error=((opt_h)&&(optind!=argc)))) {
    cerr << "No positional aruments allowed."
	 << endl;
  }
  if ((!error)&&(error=((!opt_h)&&(!opt_i)&&(optind!=argc-1)))) {
    cerr << "There must be exactly one non-option filename argument, the yacts script."<< endl;
  }
  if ((!error)&&(error=(opt_i&&(!((optind==argc)||(optind==argc-1)))))) {
    cerr << "Interactive mode may specify the filename argument, the yacts script."<< endl;
  };
  if (error) {
    cerr << "Please use -h option to see help or check the man page." << endl;
    exit(9);
  }
  if (opt_h) {
    char help[] = {
#include "yacts_help.h"
    };
    cout << "Welcome to " PACKAGE_STRING 
	 << " : yet another continuous time simulator." << endl 
	 << endl << help 
	 << "For additional details see the man page." << endl;
    exit(0);
  }
  jplus::initJlibrary(opt_i?cout:cerr);
  yacts ex;
  if (!ex.setOut(out)) {
    cerr << "Invalid output variable name \""<<out<<"\"."<<endl;
    exit(9);
  };
  string prefix="";
  if ((opt_i)&&(optind==argc)) {  // no filename given
      istringstream script(""); // empty script 
      ex.init(script);
    } 
  else if ((argv[optind][0]=='-')&&(argv[optind][1]==0))
    ex.init(cin);
  else {
    prefix=argv[optind];
    ifstream in(argv[optind]);
    if (!in) {
      cerr << "Error opening script file: " << argv[optind] << " ." << endl;
      exit(9);
    };
    ex.init(in);
  };
  if (opt_i)
    return ex.REPL();
  if (opt_p) { // processing only
    ex.initTrajectory(prefix);
    if (fn<0) fn=ex.size(); // number of frames
    // now we divide fn frames into nj chunks and process the chunk numbered nc
    int perblock=round((fn*1.0)/(nj*1.0));
    int fnstart=nc*perblock;
    int fnend=min((nc+1)*perblock,fn);
    int i=fnstart;
    bool loaded_ok=ex.setFrame(fnstart);
    assert(loaded_ok);
    while (i<fnend) {
      cout << ex.process() << endl;
      if (i<(fnend-1)) {
	loaded_ok=ex.nextFrame();
	assert(loaded_ok);
      };
      i++;
    };
    if ((nc==(nj-1))&&opt_f) { // if last block and we follow
      while (true) {
	int nsize=ex.size();
	while (i<nsize) {
	  loaded_ok=ex.nextFrame();
	  assert(loaded_ok);
	  cout << ex.process() << endl;
	  i++;
	};
	WAIT;
      };
    };
    exit(0);
  };
  // we go to the last frame and simulate until the end
  // writing the trajectory and processing if requested
  ex.initTrajectory(prefix);
  if ((!opt_s)&&(ex.size()==1)) cout << ex.process() << endl;
  while (ex.nextFrame()) {
    if (!opt_s) cout << ex.process() << endl;
    // for (int i=0; i<100; i++) { WAIT; };
  };
  exit(0);
}
